unit uHashList;
{$INCLUDE 'vfy.inc'}
{.$DEFINE DEBUG_HASHLIST_LOAD_SPEED}

interface

uses
  SysUtils, Classes, Windows, Contnrs,
  {$IFDEF DELPHIXE4_UP} Types, {$ENDIF}
  uHashItem, uProgress, uFiles, uStrings;

type
  THashList = class
    private
      fList: TObjectList;
      function GetItem(Index: integer): THashItem;
      procedure SetItem(Index: integer; Value: THashItem);
      function GetCount: integer;
    protected
      function IsSorted(const Compare: TListSortCompare): boolean;
      property List: TObjectList read fList;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Clear;
      procedure Delete(Index: integer);
      function Add(Item: THashItem): Integer;
      function AddCopy(Item: THashItem): Integer;
      procedure Insert(Index: Integer; Item: THashItem);
      procedure LoadFromFile(const FileName: string; const BasePath: string = ''; const OnProgress: TProgressEvent = nil; const DefaultState: THashState = hsOK);
      procedure SaveToFile(const FileName: string; const OnProgress: TProgressEvent = nil; const SkipStates: THashStates = []);
      procedure DeleteByState(State: THashState);
      procedure SortByPath;
      function FindByPath(const Path: string; out Index: integer): boolean; // Requires HashList sorted by Path
      procedure SortByHash;
      function FindByHash(const Hash: THashValue; const Size: int64; out Index: integer): boolean; // Requires HashList sorted by Hash
      property Items[Index: Integer]: THashItem read GetItem write SetItem; default;
      property Count: integer read GetCount;
    end;

implementation

{ THashList }

constructor THashList.Create;
begin
  inherited Create;
  fList := TObjectList.Create(True);
end;

destructor THashList.Destroy;
begin
  FreeAndNil(fList);
  inherited;
end;

function THashList.GetItem(Index: integer): THashItem;
begin
  Result := THashItem(List.Items[Index]);
end;

procedure THashList.SetItem(Index: integer; Value: THashItem);
begin
  List.Items[Index] := Value;
end;

function THashList.Add(Item: THashItem): Integer;
begin
  Result := List.Add(Item);
end;

function THashList.AddCopy(Item: THashItem): Integer;
var
  ItemCopy: THashItem;
begin
  ItemCopy := THashItem.Create;
  try
    ItemCopy.Assign(Item);
    Result := Add(ItemCopy);
  except
    FreeAndNil(ItemCopy);
    Raise;
  end;
end;

procedure THashList.Insert(Index: Integer; Item: THashItem);
begin
  List.Insert(Index, Item);
end;

function THashList.IsSorted(const Compare: TListSortCompare): boolean;
var
  Last: THashItem;
  i: integer;
begin
  Result := True;
  if Count > 1 then
    begin
    Last := Items[0];
    for i := 1 to Pred(Count) do
      if Compare(Last, Items[i]) > 0 then
      begin
        Result := False;
        Break;
      end
      else
        Last := Items[i];
    end;
end;

procedure THashList.LoadFromFile(const FileName, BasePath: string; const OnProgress: TProgressEvent; const DefaultState: THashState);
const
  MAX_FILE_SIZE_FOR_DIRECT_READ = 256*1024*1024;
var
  s: string;
  Item: THashItem;
  Done, Total, LineNr: int64;
  StdIn: boolean;
  Stream: TTextStream;
  {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
  TicksReading: DWORD;
  TicksParsing: DWORD;
  TicksList: DWORD;
  TicksTotal: DWORD;
  {$ENDIF}

  function DoReadLine(out Line: string): boolean;
    begin
      {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
      TicksReading := TicksReading - GetTickCount;
      {$ENDIF}
      if StdIn then
      begin
        Result := not Eof(Input);
        if Result then
          Readln(Input, Line);
      end
      else
      begin
        Result := Stream.ReadLine(Line);
      end;
      {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
      TicksReading := TicksReading + GetTickCount;
      {$ENDIF}
    end;

begin
  {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
  TicksTotal := -GetTickCount;
  TicksReading := 0;
  TicksParsing := 0;
  TicksList := 0;
  {$ENDIF}
  StdIn := FileName = '';
  if StdIn then
    Stream := nil
  else if THashItem.GetFileSize(FileName) <= MAX_FILE_SIZE_FOR_DIRECT_READ then
    Stream := TTextStream.Create( {$IFDEF UNICODE} LoadFileIntoMemoryW {$ELSE} LoadFileIntoMemoryA {$ENDIF} (FileName, feAuto), True)
  else
    Stream := TTextStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    Done := 0;
    LineNr := 0;
    if StdIn then
      Total := -1
    else
      Total := Stream.Size;
    while DoReadLine(s) do
    begin
      Inc(LineNr);
      Inc(Done, Length(s)+2);
      s := Trim(s);
      if s <> '' then
        if s[1] <> ';' then
        begin
          Item := THashItem.Create;
          try
            Item.BasePath := BasePath;
            Item.State := DefaultState;
            {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
            TicksParsing := TicksParsing - GetTickCount;
            {$ENDIF}
            Item.AsString := s;
            {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
            TicksParsing := TicksParsing + GetTickCount;
            TicksList := TicksList - GetTickCount;
            {$ENDIF}
            Add(Item);
            {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
            TicksList := TicksList + GetTickCount;
            {$ENDIF}
          except
            on E: EHashItemError do
            begin
              E.Message := Format('Line %d - %s', [LineNr, E.Message]);
              Item.Free;
              Raise;
            end;
            on Exception do
            begin
              Item.Free;
              Raise;
            end;
          end;
        end;
      if Assigned(OnProgress) and (Total >= 0) then
        OnProgress(Self, Done, Total);
    end;
  finally
    FreeAndNil(Stream);
  end;
  {$IFDEF DEBUG_HASHLIST_LOAD_SPEED}
  TicksTotal := TicksTotal + GetTickCount;
  Writeln(Format('THashList.LoadFromFile: Total time = %d.%03.3ds', [TicksTotal div 1000, TicksTotal mod 1000]));
  Writeln(Format('THashList.LoadFromFile: Reading time = %d.%03.3ds', [TicksReading div 1000, TicksReading mod 1000]));
  Writeln(Format('THashList.LoadFromFile: Parsing time = %d.%03.3ds', [TicksParsing div 1000, TicksParsing mod 1000]));
  Writeln(Format('THashList.LoadFromFile: List time = %d.%03.3ds', [TicksList div 1000, TicksList mod 1000]));
  Abort;
  {$ENDIF}
end;

procedure THashList.SaveToFile(const FileName: string; const OnProgress: TProgressEvent; const SkipStates: THashStates);
var
  s: string;
  i: integer;
  StdOut: boolean;
  Stream: TTextStream;
begin
  StdOut := FileName = '';
  if StdOut then
    Stream := nil
  else
    Stream := TTextStream.Create(FileName, fmCreate);
  try
    if not StdOut then
      Stream.Encoding := feUtf8;
    for i := 0 to Pred(Count) do
    begin
      if not (Items[i].State in SkipStates) then
      begin
        s := Items[i].AsString;
        if StdOut then
          Writeln(s)
        else
          Stream.WriteLine(s);
        end;
      if Assigned(OnProgress) then
        OnProgress(Self, Succ(i), Count);
    end;
  finally
    FreeAndNil(Stream);
    end;
end;

procedure THashList.SortByPath;
begin
  if not IsSorted(HashItemCompareByPath) then
    List.Sort(HashItemCompareByPath);
end;

function THashList.FindByPath(const Path: string; out Index: integer): boolean;
var
  LowIndex, HighIndex, TestIndex, CompareResult: Integer;
begin
  Result := False;
  LowIndex := 0;
  HighIndex := Pred(Count);
  while LowIndex <= HighIndex do
  begin
    TestIndex := (LowIndex + HighIndex) shr 1;
    CompareResult := AnsiCompareText(Items[TestIndex].Path, Path);
    if CompareResult < 0 then
      LowIndex := Succ(TestIndex)
    else
    begin
      HighIndex := Pred(TestIndex);
      if CompareResult = 0 then
        Result := True;
    end;
  end;
  Index := LowIndex;
end;

procedure THashList.SortByHash;
begin
  if not IsSorted(HashItemCompareByHash) then
    List.Sort(HashItemCompareByHash);
end;

function THashList.FindByHash(const Hash: THashValue; const Size: int64; out Index: integer): boolean;
var
  LowIndex, HighIndex, TestIndex, CompareResult: Integer;
begin
  Result := False;
  LowIndex := 0;
  HighIndex := Pred(Count);
  while LowIndex <= HighIndex do
  begin
    TestIndex := (LowIndex + HighIndex) shr 1;
    if Items[TestIndex].Hash < Hash then
      CompareResult := -1
    else if Items[TestIndex].Hash > Hash then
      CompareResult := 1
    else if Items[TestIndex].Size < Size then
      CompareResult := -1
    else if Items[TestIndex].Size > Size then
      CompareResult := 1
    else
      CompareResult := 0;
    if CompareResult < 0 then
      LowIndex := Succ(TestIndex)
    else
    begin
      HighIndex := Pred(TestIndex);
      if CompareResult = 0 then
        Result := True;
    end;
  end;
  Index := LowIndex;
end;

procedure THashList.DeleteByState(State: THashState);
var
  i: integer;
begin
  for i := Pred(Count) downto 0 do
    if Items[i].State = State then
      List.Delete(i);
end;

function THashList.GetCount: integer;
begin
  Result := List.Count;
end;

procedure THashList.Clear;
begin
  List.Clear;
end;

procedure THashList.Delete(Index: integer);
begin
  List.Delete(Index);
end;

end.
